                  /*********************************/
                  /*              NOTE             */
                  /*                               */
                  /* This sample code has been     */
                  /* provided by IBM.  It is not   */
                  /* warranted for any particular  */
                  /* use or purpose.               */
                  /*                               */
                  /* IBM releases this code into   */
                  /* the public domain.  You may   */
                  /* use it, modify it, or         */
                  /* incorporate it into other     */
                  /* products without restriction. */
                  /*********************************/
 /*********************************************************************/
 /*                                                                   */
 /* MODULE  NAME:    rfio.c                                           */
 /*                                                                   */
 /* DESCRIPTIVE NAME: Low-level routines for Response File I/O        */
 /*                                                                   */
 /* LINKAGE:   These are linkable subroutines. They require           */
 /*            CID_findfile()   be linked in also.                    */
 /*                                                                   */
 /*********************************************************************/
#pragma page (1)
/* rfio.c -- contains Response File I/O APIs */
/* Author: Bob Blair */

/* Contents:

NOTE : when the DTCO routines are substituted for the strxxx calls in these
       routines, they are DBCS enabled.  See rfioi.h

******   File Open and Close *******

RFHANDLE RF_Open_File (
                              char *filename,   // The name of the resp file
                              RFISP root_isp    // Pointer to the root of the
                                                // include stack
                              int  *DosRc       // Where to put DOS return code
                           );

       RF_Open_File returns a handle that can be used with the other RF
     commands on success, or (RFHANDLE)NULL when there is an error.
     In the case of an error, errno is put at *DosRc;
     DosRc is unchanged if the operation is successfull.

       filename points to an asciiz string containing the path and name of
     the response file.  RF_Open_File does not presume a particular
     file suffix, so you have to give the whole thing  i.e. \sd\myrf.rsp
     RF_Open_File first tries DosFindFirst on the file spec.  If it is
     not found, and the filespec is relative, then each unit of DPATH is
     searched.  If still not found, each unit of PATH is searched.
     The file name is allowed to be ambiguous, but only the first occurance
     found is returned.

       root_isp is a pointer to a structure of type RFIS, which is used to
     handle the imbedding of included files.  A call to RF_Open_File
     preserves the handle of the file previously opened by RF_Open_File
     when it returns the new handle.  RF_Close_File returns the previous
     handle, it there was one.  So RF_Open and RF_Close can be used together
     to painlessly process imbedded references to other files.

       RF_Open_File checks for include file loops.  If an open is issued
     for file that is already in the include stack,  it will be ignored;
     the current RFHANDLE will be returned with no error indication.

       The question arises what should the caller do if she is trying to
     open an imbed file and NULL (file not found) is returned.  Normally,
     this will be a fatal error.  If the caller wants to recover and
     resume processing, however, she can call RF_Get_Current_RFHANDLE()
     to get the handle of the file that included the imbed; then resume
     processing.


RFHANDLE RF_Close_File     (
                              RFHANDLE rfh    // Value returned from open
                              RFISP root_isp  // Pointer to root of include
                                              //   stack
                           );

       RF_Close_File closes the response file.

       rfh is the RFHANDLE to be close.  Close looks in the include stack,
     and closes the file associated with rfh and all files below it on
     the stack (normally, rfh will probably be at the bottom of the
     stack).

       root_isp is a pointer to the root entry of the include stack.

       RF_Close_File returns the RFHANDLE that was in effect when rfh
     was opened, or (RFHANDLE)NULL if rfh was the first file opened.
/*
#pragma page (1)
/*
*******  Routines for manipulating the include stack  *********

RFISP RF_Create_Root_ISE (void);

       This routine creates the root imbed (a.k.a. include) stack entry.
     A user of RFIO calls it exactly once, before the exection of the
     first RF_Open_File call.  It returns a value passed to all subsequent
     Open and Close calls.  On an out-of-memory condition, it returns
     (RFISP)NULL.

char *RF_Get_Current_File_Name(
                              RFISP root_isp;
                              );

        This routine returns a pointer to the full name by which the
      currently active response file was found.  You need to free()
      the string after you use it.

        If there is no response file currently open, or if there is an
      out-of-memory condition, NULL is returned.

char *RF_Get_Current_RFHANDLE(
                             RFISP root_isp;
                             );

        This routine returns the RFHANDLE of the currently active
      response file.

        If there is no response file currently open, NULL is returned.

/*
#pragma page (1)
/*
*******  Sequential File I/O *********

int RF_Get_Next_Kwd_In_File (
                              RFHANDLE rfh,  // Value returned from open
                              char **kwp,    // Where to put the pointer to
                                             // the key word.
                              char **valp,   // Where to put the pointer to
                                             // the value.
                              int *valtypep  // Where to put the type of value
                                 );          // (0 = string; 1 = list)

       This function reads the next response file entry, parses it, creates
     a place to put the pieces, and puts the pieces there.  It returns
     pointers to the keyword string and the value string; both the strings
     are null-terminated.  Comment lines and blank lines are skipped.
     In the case of a list, a pointer to the whole list is returned.

       rfh is the handle returned from the RF_Open_File call.

       kwp is the address of a character pointer where the routine will put
     the address of the keyword string.

       valp is the address of a character pointer where the routine will put
     the address of the value string.

       valtypep is a pointer to an integer where the routine will put the type
     of value returned.  0 if a simple value string, or 1 if a list.

     Returns
         RF0          to indicate success
         RFOOS        to indicate an out-of-memory condition
         RFEOF        to indicate end-of-file
         RFERR        if their is an I/O error or other DOS error.
                      The OS error code is stuffed at *valtypep.
         RFSYNTAX     to indicate a Response File syntax error
                      You can get the file name with a call to
                      RF_Get_Current_File_Name().
                      The offending line at *valp, and the line number at *type.

int RF_Get_Next_Kwd_In_List(
                              char **start,  // val returned from
                                             // RF_Get_Next_Kwd_In_File()
                              char **kwp,    // Where to put the pointer to
                                             // the key word.
                              char **valp,   // Where to put the pointer to
                                             // the value.
                              int *valtypep  // Where to put the type of value
                                 );          // (0 = string; 1 = list)

       This is like RF_Get_Next_Kwd_In_File() except that it works on lists
     in memory.  Instead of an RFHANDLE, it takes a double char pointer that
     keeps track of where it is in the list.
     It is used to sequentially parse list values returned from
     RF_Get_Next_Kwd_In_File().

       start is a pointer to a char pointer.  On
     the first invocation, the caller must set it to the valp parameter
     returned by RF_Get_Next_Kwd_In_File().  On subsequent calls, it is a value
     set by this routine.  The caller should not modify this value after
     setting it to valp.

       Note that nested lists are supported.

       This routine returns the same values as RF_Get_Next_Keyword_In_List.
     When RFEOF is returned, it really means that end-of-list has been
     reached.

int RF_Get_Next_Wanted_Kwd_In_File(
                              RFHANDLE rfh,  // Value returned from open
                              char **kwdlist,// A pointer to an array of
                                             // character pointers.  Each
                                             // element of the array points to
                                             // a string that contains the name
                                             // of a keyword of interest.
                              char **kwp,    // Where to put the pointer to
                                             // the key word.
                              char **valp,   // Where to put the pointer to
                                             // the value.
                              int *valtypep  // Where to put the type of value
                              );             // (0 = string; 1 = list)

        This routine works just lie RF_Get_Next_Kwd_In_File() except that you
     can restrict the keywords returned to those you specify.  The keywords
     to be returned are pointed to by kwdlist.  This variable should point
     to a vector of pointers to strings that contains the names of keywords
     of interest.  The strings must be asciiz;  the last element in the vector
     of pointers must be NULL.

        Here's an example:
     char *kwdvector[] = {
                           "kwd1",
                           "keyword2",
                           "kword3",
                           NULL
                         }
     while (RF_Get_Next_Wanted_Kwd_In_File(rfh, kwdvector, kwp, valp, typep)
     {                                                                  != RFEOF)
       <process your keyword>
       free(kwp);
       free(valp);
     }


       This routine returnes the same values as RF_Get_Next_Wanted_Kwd_In_File.
     RFEOF means the last matching keyword has already been returned.

/*
#pragma page (1)
/*
*******  Sample Usage *******

RFHANDLE rfh;
RFISP isp;
char *filename, *kw, *val;
int rc, type;

if ( (rfisp = RF_Create_Root_ISE()) )
  if ( (rfh = RF_Open_File(filename, rfisp, &rc)) )
    while (rfh)
      {
        rc = RF_Get_Next_Kwd_In_File(rfh, &kw, &val, &type)
        switch (rc)
        {
          case RFSYNTAX:           <log the error; line number is at type,
                                    the bad line is at val.>
                                   break;

          case RFOOS:              <Log the error.  It's probably best to
                                    always treat this as a fatal error.
                                   break;

          case RFERR:              <Log the error.  This is probably a fatal
                                    error, too.>
                                   break;

          case RFEOF:              rfh = RF_Close_File(rfisp);
                                   break;

          case RF0:                if (type == RFLIST)
                                     <process the list using
                                      RF_Get_Next_Kwd_In_List() >
                                   else  // type == RFSTRING
                                     <if this is an Include>
                                        rfh = RF_Open_File(val, rfisp, &rc);
                                      else
                                        <process kw = val>
                                   break;
        } // end SWITCH
      } // end WHILE

*/

#pragma page (1)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <ctype.h>         /* for isspace().   May not be needed always. */
#include <errno.h>
#define INCL_BASE
#include <os2.h>
#include "rfioi.h"


/* **** Private routines ***** */
RFISP    RF_get_last_ise(RFISP isp);
RFHANDLE RF_get_last_stream(RFISP isp);
int      RF_find_imbed_loop(RFISP isp, char *name);
char    *RF_find_file(char *fn);
RFISP    RF_create_imbed_stack_entry(RFISP isp);
int      RF_get_list(RFHANDLE rfh, char *buff, char **val, int *type);
int      RF_parse_RF_line(char *line, char **kw, char **val);
char    *RF_lastnonwhite(char *start);      /* NOTE: these three routines  */
char    *RF_firstwhiteorequal(char *start); /* will be different depending */
char    *RF_firstnonwhite(char *start);     /* on DBCS enabling method     */

/* **** External routine **** */
extern char    *CID_findfile(char *fn);
#pragma page (1)

/* *********************************************************************** */
/*                           RF_Open_File                                  */
/* *********************************************************************** */
RFHANDLE RF_Open_File(char *filename, RFISP isp, int *rcp)
{
    FILE *stream;
    char *fqname;
    RFISP ise;

    stream = (FILE *)NULL;

    if (rcp != (int *)NULL)
//      if ((isp == (RFISP *)NULL) || (filename == (char *)NULL))      @A1d
      if ((isp == (RFISP)NULL) || (filename == (char *)NULL))        //@A1a
         *rcp = RF_BAD_CALL;
      else
        if ( (fqname = RF_find_file(filename)) == NULL )
          *rcp = RF_FILE_NOT_FOUND;
        else
          if (RF_find_imbed_loop(isp, fqname))
            {
               free(fqname);
               stream = RF_get_last_stream(isp);
            }
          else
            if ( (stream = fopen(fqname, RF_RDONLY)) == (FILE *)NULL)
              *rcp = errno;
            else
              {
                ise = RF_create_imbed_stack_entry(isp);
                if (ise)
                  {
                    ise->rfis_rfh = stream;
                    ise->rfis_filename = fqname;
                    ise->rfis_previous->rfis_lineno = RF_lineno;
                    RF_lineno = 0;
                  }
              }
    return(stream);
}
#pragma page (1)

/* *********************************************************************** */
/*                           RF_Close_File                                 */
/* *********************************************************************** */
RFHANDLE RF_Close_File(RFHANDLE current_handle, RFISP isp)
{
    RFISP myisp, disp, dispprev;

    if (isp == (RFISP)NULL)
      return(current_handle);

    if (current_handle == (RFHANDLE)NULL)
      return(RF_get_last_stream(isp));

    for (myisp = isp;
           myisp->rfis_next && myisp->rfis_rfh != current_handle;
             myisp = myisp->rfis_next);
              /* Close ANY files that were opened after the one being closed */
    if (myisp->rfis_rfh == current_handle)
      {
        disp = RF_get_last_ise(isp);
        dispprev = (RFISP)NULL;
        while(disp)
          {
            dispprev = disp->rfis_previous;
            fclose((FILE *)disp->rfis_rfh);
            free(disp->rfis_filename);
            free(disp);
            if (disp != myisp)
              disp = dispprev;
            else
              disp = (RFISP)NULL;
          }
        if (dispprev)
          {
            dispprev->rfis_next = (RFISP)NULL;
            RF_lineno = dispprev->rfis_lineno;
            return(dispprev->rfis_rfh);
          }
        else
          return((RFHANDLE)NULL);                    //@A1c
      }
    else       /* We were handed a handle we don't know about */
      {
         fclose(current_handle);    /* try to close it anyway */
         dispprev = RF_get_last_ise(isp);  /* return last one open */
         RF_lineno = dispprev->rfis_lineno;
         return(RF_get_last_stream(isp));  /* return last one open */
      }
}
#pragma page (1)


/* *********************************************************************** */
/*                        RF_Create_Root_ISE                               */
/* *********************************************************************** */
RFISP RF_Create_Root_ISE(void)
{

    return( RF_create_imbed_stack_entry((RFISP)NULL) );
}
#pragma page (1)


/* *********************************************************************** */
/*                      RF_Get_Current_Filename                            */
/* *********************************************************************** */
char *RF_Get_Current_Filename(RFISP isp)
{
   RFISP ise;
   char *p;

   ise = RF_get_last_ise(isp);
   if (!ise)
     return(NULL);
   p = (char *)malloc(strlen(ise->rfis_filename) + CHARSIZE);
   if (p)
     strcpy(p, ise->rfis_filename);
   return(p);
}
#pragma page (1)

/* *********************************************************************** */
/*                      RF_Get_Current_RFHANDLE                            */
/* *********************************************************************** */
RFHANDLE RF_Get_Current_RFHANDLE(RFISP isp)
{
   RFISP ise;

   ise = RF_get_last_ise(isp);
   if (!ise)
     return(NULL);
   return(ise->rfis_rfh);
}
#pragma page (1)

/* *********************************************************************** */
/*                     RF_Get_Next_Kwd_In_File                             */
/* *********************************************************************** */
int RF_Get_Next_Kwd_In_File(RFHANDLE rfh, char **kw, char **val, int *type)
{
   int rc, lrc;
   char linebuff[RF_MAXLINE + 1];

   while (FOREVER)
     {
       if (fgets(linebuff, RF_MAXLINE, (FILE *)rfh) == NULL)
         {
            if (!feof((FILE *)rfh))
              {
                 *type = errno;
                 return(RFERR);
              }
            else
              return(RFEOF);
         }
       RF_lineno++;
       rc = RF_parse_RF_line(linebuff, kw, val);
       switch (rc)
       {
           case RFOOS:           return(rc);
                                 break;

           case RFKWONLY:
           case RFSYNTAX:
                                 *val = (char *)malloc(strlen(linebuff) + 1);
                                 strcpy(*val, linebuff);
                                 *type = RF_lineno;
                                 return(RFSYNTAX);
                                 break;

           case RF0:             *type = RFSTRING;
                                 return(rc);
                                 break;

           case RFLIST:          *val = NULL;
                                 lrc = RF_get_list(rfh, linebuff, val, type);
                                 if (lrc == RF0)
                                   *type = RFLIST;
                                 return(lrc);
                                 break;
       }
     }

}
#pragma page (1)
/* *********************************************************************** */
/*                     RF_Get_Next_Kwd_In_List                             */
/* *********************************************************************** */
int RF_Get_Next_Kwd_In_List(char **start, char **kw, char **val, int *type)
{
   char *p, *q, *mykw, *myval;
   int rc, ildepth, llen;

   ildepth = 0;
   if (start == NULL || *start == NULL || **start == '\0')
     {
       *type = RFEOF;
       return(RFEOF);
     }
   rc = RF_parse_RF_line(*start, kw, val);
       switch (rc)
       {
           case RFOOS:           return(RFOOS);
                                 break;

           case RFSYNTAX:        free(*val);
                                 *val = (char *)malloc(strlen(*start) + CHARSIZE);
                                 strcpy(*val, *start);
                                 *type = RF_lineno;
                                 return(RFSYNTAX);
                                 break;

           case RFKWONLY:
           case RF0:             *type = RFSTRING;
                                 *start = *start + strlen(*start) + CHARSIZE;
                                 return(RF0);
                                 break;

           case RFLIST:          *val = NULL;
                                 *type = RFLIST;
                                 p = *start + strlen(*start) + CHARSIZE;
                                 for (q = p; *q; q += strlen(q) + CHARSIZE)
                                   {
                                     if (RF_parse_RF_line(q, &mykw, &myval)
                                                                      == RFLIST)
                                       ildepth++;
                                     else
                                       if (strcmp(mykw, RF_LISTENDS) == 0)
                                         if (ildepth)
                                           ildepth--;
                                         else
                                           {
                                             free(mykw);
                                             free(myval);
                                             break;
                                           }
                                     free(mykw);
                                     free(myval);
                                   }
                                 if (!*q)
                                   {
                                     free(*val);
                                     *val = (char *)malloc(strlen(*start) + CHARSIZE);
                                     strcpy(*val, *start);
                                     *type = RF_lineno;
                                     return(RFSYNTAX);
                                   }
                                 else
                                   {
                                     llen = (q-p);
                                     *val = malloc( llen + CHARSIZE);
                                     if (*val)
                                       {    /* this copy copies the last NULL */
                                         memcpy((void *)*val,
                                                   (void *)p,
                                                       llen);
                                         p = *val + llen;
                                            /* second null terminates list */
                                         *p = '\0';
                                       }
                                     *start = *start + strlen(*start)
                                            + llen   + strlen(q)
                                            + CHARLENT2;
                                   }
                                 return(RF0);
                                 break;
       }

}
#pragma page (1)


/* *********************************************************************** */
/*                     RF_Get_Next_Wanted_Kwd_In_File                      */
/* *********************************************************************** */
int RF_Get_Next_Wanted_Kwd_In_File(RFHANDLE rfh, char **kwdlist, char **kw,
                                   char **val, int *type)
{
    char *mykw, *myval;
    int mytype, i, rc;

    while ((rc=RF_Get_Next_Kwd_In_File(rfh, &mykw, &myval, &mytype)) != RFEOF)
      {
        for (i = 0; kwdlist[i]; i++)
           if (stricmp(mykw, kwdlist[i]) == 0)
             {
               *kw = mykw;
               *val = myval;
               *type = mytype;
               return(rc);
             }
      }
    return(RFEOF);
}
#pragma page (1)

/* ************************** PRIVATE ROUTINES ************************** */

/**************************************************************************
*    RF_create_imbed_stack_entry creates an instance of structure RFIS.   *
*                                                                         *
*    It initializes the structure, and, unless the root pointer (passed   *
*    as isp) is NULL, puts it to the end of the imbed stack.              *
*                                                                         *
*    Call this routine with isp = NULL to create the root stack entry.    *
*    Call this routine with isp = <the root stack entry> to create        *
*    and add a new entry to the stack.                                    *
**************************************************************************/

RFISP RF_create_imbed_stack_entry(RFISP isp)  /* we are passed the root */
{
   RFISP ise, newise;

   if ( (newise = (RFISP)malloc(sizeof(RFIS))) != NULL)
     {
       newise->rfis_rfh = (RFHANDLE)NULL;
       newise->rfis_filename = (char *)NULL;
       newise->rfis_lineno = 0;
       newise->rfis_next = (RFISP)NULL;
       if (isp)
         {
           for (ise = isp; ise->rfis_next; ise = ise->rfis_next);
           ise->rfis_next = newise;
           newise->rfis_previous = ise;
         }
       else
         newise->rfis_previous = (RFISP)NULL;
       return(newise);
     }
   else
     return((RFISP)NULL);
}
#pragma page (1)

/**************************************************************************
*    RF_find_imbed loop looks through the stack of imbedded files         *
*    to try to find a name that duplicates the name of a file about       *
*    to be imbedded.  If the names match, then a loop condition would     *
*    result, since Response Files are processed sequentially with no      *
*    logic.                                                               *
*                                                                         *
*    This routine can't guarantee to catch all loop conditions.  If the   *
*    response file contains an imbed name in two different forms that     *
*    resolve to the same file (i.e. C:filename   and currdrive\filename)  *
*    then this code will not catch it.  It does provide some protection,  *
*    though.                                                              *
**************************************************************************/

int RF_find_imbed_loop(RFISP isp, char *name)
{
   RFISP ise;

   if (isp)
     for (ise = isp->rfis_next; ise; ise = ise->rfis_next)
        if (stricmp(name, ise->rfis_filename) == 0)
          return(1);

   return(0);
}
#pragma page (1)

/**************************************************************************
*    RF_get_last_stream returns the FILE pointer associated with the      *
*    most recent stack entry.  It is mainly used in error situations.     *
*    isp is the root of the imbed stack.                                  *
**************************************************************************/
RFHANDLE RF_get_last_stream(RFISP isp)
{
   RFISP ise;

   for (ise = isp; ise->rfis_next; ise = ise->rfis_next);
   return(ise->rfis_rfh);
}
#pragma page (1)

/**************************************************************************
*    RF_get_last_ise returns a pointer to the most recent stack entry.    *
*    isp is the root of the imbed stack.                                  *
**************************************************************************/
RFISP RF_get_last_ise(RFISP isp)
{
   RFISP ise;

   for (ise = isp; ise->rfis_next; ise = ise->rfis_next);
   return(ise);
}
#pragma page (1)

/**************************************************************************
* Given a file spec, find the first file that satisfies it.               *
*                                                                         *
*    The algorithem is:  look on the current directory.                   *
*    If it's not there, and the filespec was fully qualified, return      *
*    file-not-found.                                                      *
*    If it's not there but the filespec was relative, look up the DPATH   *
*    concatenate the file with each directory listed there, and search    *
*    for the resultant fully qualified name.                              *
*    If it's still not found, look up the PATH and search on each         *
*    directory listed there.                                              *
*                                                                         *
*    The file spec can be ambiguous.  Only the first file that satisfies  *
*    the spec will be returned.                                           *
*                                                                         *
*    Returns a pointer to the fully-qualified file name if successful,    *
*    or NULL if the file was not found.  The reason the file was not      *
*    found is not returned.  It could be because the file does not exist  *
*    or because an I/O error prevented access.  It could also be that     *
*    there is not enough memory left to store the file name.              *
**************************************************************************/

char *RF_find_file(char *filespec)
{
    return(CID_findfile(filespec));
}
#pragma page (1)


/**************************************************************************
*   RF_get_list reads in each item of a list, trims off white space, and  *
*   stores it in a vector.  List items are seperated by nulls.  End of    *
*   list is marked by two consecutive nulls.                              *
**************************************************************************/
int RF_get_list(RFHANDLE rfh, char *buff, char **val, int *type)
{
   char *mykw, *myval, *newval, *p;
   int rc, len, ildepth;
   size_t valsize, oldsize;

   ildepth = 0;      /* this parm tracks nested lists */
   if ( (*val = (char *)malloc(CHARSIZE)) == NULL)
     return(RFOOS);
   valsize = 0;
   **val = '\0';

   while (FOREVER)
     {
       if (fgets(buff, RF_MAXLINE, (FILE *)rfh) == NULL)
         {
            if (!feof((FILE *)rfh))
              {
                 *type = errno;
                 return(RFERR);
              }
            else
              return(RFSYNTAX);
         }
       RF_lineno++;
       rc = RF_parse_RF_line(buff, &mykw, &myval);
       switch (rc)
       {
           case RFOOS:           return(rc);
                                 break;

           case RFSYNTAX:        free(*val);
                                 *val = (char *)malloc(strlen(buff) + 1);
                                 strcpy(*val, buff);
                                 *type = RF_lineno;
                                 return(RFSYNTAX);
                                 break;

           case RFKWONLY:        if (strcmp(mykw, RF_LISTENDS) == 0)
                                   if (ildepth == 0)
                                     {
                                       p = *val + valsize;
                                       *p = '\0'; // second null
                                       return(RF0);
                                     }
                                   else
                                     ildepth--;
                                 else
                                   if (strcmp(mykw, RF_LISTSTARTS) == 0)
                                     ildepth++;
                         /* No break! fall through */
           case RFLIST:
           case RF0:             len = strlen(mykw);
                                 oldsize = valsize;
                                 len += CHARLENT2;
                                 if (myval && *myval)
                                   {
                                      len += strlen(myval);
                                      len += strlen("=");
                                            /* we put an extra byte in here so*/
                                            /* we won't have to allocate an   */
                                            /* extra byte to mark the list's  */
                                            /* end                            */
                                   }
                                 valsize += len;
                                 newval = (char *)realloc((void *)*val, valsize);
                                 if (newval == NULL)
                                   return(RFOOS);
                                 else
                                   {
                                     p = newval + oldsize;
                                     *val = newval;
                                     strcpy(p, mykw);
                                     if (myval && *myval)
                                       {
                                         strcat(p, "=");
                                         strcat(p, myval);
                                       }
                                        /* here we put the real data size into*/
                                        /* valsize.  Later I can change       */
                                        /* this line to valsize-=CHARSIZE     */
                                     valsize = oldsize + strlen(p) + CHARSIZE;
                                   }
                                 if (myval && *myval)
                                   if (strcmp(myval, RF_LISTSTARTS) == 0)
                                     ildepth++;
                                 break;
       }
     }
}
#pragma page (1)
/**************************************************************************
*   Take a string  presumed to be in the form                             *
*                     keyword    =    value       \n                      *
*   where the white space is optional.  If the first non-white character  *
*   is an asterisk, the line is a comment.  If the line is all blank,     *
*   it is a comment.  If the line begins with an '=', or has the form     *
*                     keyword    =                \n                      *
*   then a syntax exception is generated.                                 *
*   A line that contains just a keyword is allowed to live, but a special *
*   value RFKWONLY is returned.                                           *
**************************************************************************/

int RF_parse_RF_line(char *line, char **kw, char **val)
{
    char *p, *q, *r;

    p = RF_firstnonwhite(line);
    if (*p == '\0')
      return(RFCOMMENT);       /* Blank line:  it's a comment */
    if ( (*p == RF_COMMENTMARK1) || (*p == RF_COMMENTMARK2) )
      return(RFCOMMENT);       /* First char is '*' or ';' : it's a comment */

    q = RF_firstwhiteorequal(p);  /* This is the end of the keyword */
    if (p == q)                /* Keyword was '='                         @P1*/
      return(RFSYNTAX);           /* And that's not legal                 @P1*/
    if ( (*kw = (char *)malloc(q - p + CHARSIZE)) == NULL) /* Make space for it */
      return(RFOOS);
    strncpy(*kw, p, q-p);         /* and copy it in */
    r = *kw + (q - p);
    *r = '\0';                    /* and null-terminate it */

    p = RF_firstnonwhite(q);      /* Now look for the "=" */
    if (*p == '\0')
      {                          /* If rest of line is blank, we're done */
         *val = (char *)malloc(CHARSIZE);
         **val = '\0';
         return(RFKWONLY);
      }

    if (*p != '=')               /* If the next thing isn't '=', we have */
      return(RFSYNTAX);          /* a syntax error */

    q = RF_firstnonwhite(p+CHARSIZE);   /* If there's no value,*/
    if (*q == '\0')
      {                                                          /* @U1a */
        p = q;                                                   /* @U1a */
/*      return(RFSYNTAX);                   it's not a syntax error @U1d */
      }                                                          /* @U1a */
    else                                                         /* @U1a */
      p = RF_lastnonwhite(q);     /* Find end of value/make space for it */

    if ( (*val = (char *)malloc(p - q + CHARLENT2)) == NULL)
      return(RFOOS);
    strncpy(*val, q, p-q+CHARSIZE);   /* Copy it */
    p = *val + (p - q) + CHARSIZE;             /* null-terminate it */
    *p = '\0';
    if (strcmp(*val, RF_LISTSTARTS) == 0)    /* If this marks start of list */
      return(RFLIST);                 /* just return indication of list */
    else
      return(RF0);
}
#pragma page (1)

/**************************************************************************
*  These three routines are used to parse the response file lines.  They  *
*  need to be adapted by each product to use that product's particular    *
*  DBCS enablement mechanism.                                             *
*                                                                         *
*  The implementation below assumes a single-byte ascii character set     *
**************************************************************************/
char *RF_firstnonwhite(char *start)
{
   char *p;

   for (p = start; *p && isspace(*p); p++);
   return(p);
}

char *RF_firstwhiteorequal(char *start)
{
   char *p;

   for (p = start; *p && !isspace(*p) && *p != '='; p++);
   return(p);
}

char *RF_lastnonwhite(char *start)
{
   char *p;
   for (p = start; *p; p++);     /* find end of string */
   for (--p; *p && isspace(*p); p--);  /* work backwords */
   return(p);
}
